const si = require("systeminformation");
const os = require("os");

const PccApp = require("../../../picockpit-library/picockpit-base-app");

class Pistats extends PccApp {
  constructor() {
    super("pistats");
    this.logger.info("Constructing pistats");
    
    this.measuringLoop();
  }

  getStats = async () => {
    let stats = {};
    stats["cpu_load"] = Math.round((await si.currentLoad()).currentLoad);
    stats["ram_total_bytes"] = await os.totalmem();
    stats["ram_available_bytes"] = await os.freemem();

    const diskIo = await si.disksIO();
    const networkStats = (await si.networkStats())[0];
    const fsSize = (await si.fsSize())[0];
    const cpuTemperature = await si.cpuTemperature();
    const uptime = os.uptime();

    stats["disk_io_bytes_read"] = Math.ceil(diskIo.rIO);
    stats["disk_io_bytes_written"] = Math.ceil(diskIo.wIO);
    stats["net_io_bytes_received"] = networkStats.rx_bytes;
    stats["net_io_bytes_sent"] = networkStats.tx_bytes;
    stats["root_partition_used"] = fsSize.used;
    stats["root_partition_total"] = fsSize.size;
    stats["soc_temperature"] = Math.round(cpuTemperature.max);
    try {
      stats["wlan_signal_info"] =
        `${this.network.signalLevel} dBm, ${this.network.quality} % link quality`;
    } catch {
      stats["wlan_signal_info"] = `WLAN not enabled`;
    }
    stats["uptime"] = this.secondsToHuman(Math.floor(uptime)); // Not optimal if uptime is 1 year + --> Maybe use different library?
    return stats;
  };

  measuringLoop = async () => {
    // We only create this one, it takes a lot of time.
    this.network = (await si.wifiNetworks())[0];
    this.publishCategories();
    while (true) {
      // Don't publish too often, otherwise it is too much for Pi & Server.
      const stats = await this.getStats();
      this.logger.debug("Publishing data");
      // TODO:
      // There have been issues with the App not publishing the meta data often enough - I enable this here for now.
      // At some point this should be taken out again - preferably though when we redesign the old apps.
      this.publish("$state", stats);
      await this.sleep(2500);
    }
  };

  publishCategories = async () => {
    const meta = {
      uptime: {
        $description: "Raspberry Pi uptime",
        minutes: {
          $unit: "minute(s)",
        },
        days: {
          $unit: "day(s)",
        },
        seconds: {
          $unit: "second(s)",
        },
        hours: {
          $unit: "hour(s)",
        },
      },
      soc_temperature: {
        $description: "SoC temperature",
        $unit: "°C",
      },
      ram_total_bytes: {
        $description: "RAM total (without VideoCore)",
        $unit: "bytes",
      },
      ram_available_bytes: {
        $description: "RAM available",
        $unit: "bytes",
      },
      net_io_bytes_received: {
        $description: "Network total received data",
        $unit: "bytes",
      },
      net_io_bytes_sent: {
        $description: "Network total sent data",
        $unit: "bytes",
      },
      root_partition_used: {
        $description: "Root partition ('/') total used",
        $unit: "bytes",
      },
      root_partition_total: {
        $description: "Root partition ('/') total available",
        $unit: "bytes",
      },
      disk_io_bytes_written: {
        $description: "Disk I/O total written",
        $unit: "bytes",
      },
      disk_io_bytes_read: {
        $description: "Disk I/O total read",
        $unit: "bytes",
      },
      cpu_load: {
        $description: "CPU load (all cores)",
        $unit: "%",
      },
      wlan_signal_info: {
        $description: "WLAN signals",
        $unit: "",
      },
    };

    await this.sleep(10000);

    this.logger.verbose("Publishing $meta");
    this.publish("$meta", meta, { retain: true });
  };

  secondsToHuman = (seconds) => {
    const obj = {
      years: 0,
      days: 0,
      hours: 0,
      minutes: 0,
      seconds: 0,
    };

    if (seconds > 59) {
      obj.seconds = seconds % 60;
      obj.minutes = Math.floor(seconds / 60);
    }
    if (obj.minutes > 59) {
      obj.hours = Math.floor(obj.minutes / 60);
      obj.minutes = obj.minutes % 60;
    }
    if (obj.hours > 23) {
      obj.days = Math.floor(obj.hours / 24);
      obj.hours = obj.hours % 24;
    }
    if (obj.days > 365) {
      obj.years = Math.floor(obj.days / 365);
      obj.days = obj.days % 365;
    }

    return obj;
  };

  sleep = async (ms) => {
    return new Promise((resolve) => {
      setTimeout(resolve, ms);
    });
  };
}

pistats = new Pistats();
